/* sha - The Secure Hash Algorithm (FIPS 180)
 *
 * Implementation:  Tony Bartoletti  931204
 */

#include	<stdio.h>
#include	<string.h>

/* Algorithmic Constants */

#define		K0	0x5A827999	/* for t in 0 ..19 */
#define		K1	0x6ED9EBA1	/* for t in 20..39 */
#define		K2	0x8F1BBCDC	/* for t in 40..59 */
#define		K3	0xCA62C1D6	/* for t in 60..79 */

/* Algorithmic Global Variables */

unsigned long	H0, H1, H2, H3, H4;
unsigned long	M[16];
unsigned long	W[80];

unsigned long	S(), f0(), f1(), f2(), f3();

unsigned int	sizebuf[2];	/* accomodate 64 bit file size */
char	filename[256];

int	append_block = 0;
int	trace = 0;		/* trace output as in FIPS 180 publication */
int	blockcount = 0;
int	verbose = 0;
int	very_verbose = 0;
int	file_specified = 0;
int	more;

FILE	*GV_fp = NULL;

main(argc,argv)
int	argc;
char	*argv[];
{
	/* initialize algorithmic globals */

	init_globals();

	argv++; argc--;

	while( argc )
	{
		switch( argv[0][0] )
		{
		case '-':
			switch( argv[0][1] )
			{
			case 'h': usage_exit("");
			case 't': trace = 1;
				  break;
			case 'v': verbose = 1;
				  break;
			case 'V': very_verbose = 1;
				  break;
			default:  usage_exit("unknown switch char");
			}
			break;
		default:
			strcpy( filename, argv[0] );
			if( verbose )
				printf("Input file is %s:\n",filename);
			file_specified = 1;
			break;
		}
		argv++; argc--;
	}
	if( file_specified )
		open_file( filename );
	else
		GV_fp = stdin;

	/* load up each 512 bit (64 byte) block, and process */

	/*
	fake_it( M );
	if( trace ) show_it( M );
	process_block( M );
	print_results();
	exit(0);
	*/

	do
	{
		more = get_next_block( M, GV_fp );
		process_block( M );
	}
	while( more );

	print_results();

	if( file_specified ) fclose( GV_fp );
	exit(0);
}

init_globals()
{
	H0 = 0x67452301;
	H1 = 0xEFCDAB89;
	H2 = 0x98BADCFE;
	H3 = 0x10325476;
	H4 = 0xC3D2E1F0;

	sizebuf[0] = sizebuf[1] = 0;
}

/* stat file to insure it is a regular file, then open it */

open_file( filename )
char	*filename;
{
	char	tmp[256];

	if( (GV_fp = fopen(filename,"r")) == NULL )
	{
		sprintf(tmp,"sha: error: cannot open file %s for reading\n",
					filename);
		usage_exit(tmp);
	}
}

fake_it( M ) /* fill block M with next 64 bytes */
unsigned char	*M;
{
	int	i;

	/* message = "abc" */

	M[0] = 'a'; M[1] = 'b'; M[2] = 'c';

	/* do the padding */

	M[3] = 0x80;
	for( i=4; i<62; i++ ) M[i] = 0x00;

	M[62] = 0x00; M[63] = 0x18;
}

show_it( W ) /* print all 16 words in M */
unsigned long W[16];
{
	int	i;

	printf("\n\n");
	for( i=0; i<16; i++ )
	{
		printf("W[%d]\t= ",i);
		print_hex( W[i] );
		printf("\n");
	}
	printf("\n");
}

int
get_next_block( M, fp )
unsigned char	*M;
FILE	*fp;
{
	int	c;
	int	bytecount = 0;

	if( !append_block )
	{
		if( file_specified )
		{
			while( (c=getc(fp)) != EOF )
			{
				M[bytecount] = (unsigned char) c;
				bytecount++;
				if( bytecount == 64 ) break;
			}
		}
		else
		{
			while( (c=getc(stdin)) != EOF )
			{
				M[bytecount] = (unsigned char) c;
				bytecount++;
				if( bytecount == 64 ) break;
			}
		}
		if( bytecount == 64 )
		{
			update_sizebuf( 512 );
			return( 1 );
		}
		if( bytecount > 55 )
		{
			update_sizebuf( 8*bytecount );
			append_block = 1;
			c = bytecount;
			M[c++] = (unsigned char) 128;
			while( c < 64 ) M[c++] = (unsigned char) 0;
			return( 1 );
		}
		if( bytecount > 0 )
			update_sizebuf( 8*bytecount );
		c = bytecount;
		M[c++] = (unsigned char) 128;
		while( c < 56 ) M[c++] = (unsigned char) 0;
		stuff_file_length(M);
		return( 0 );
	}
	else
	{
		c=0;
		while( c < 56 ) M[c++] = (unsigned char) 0;
		stuff_file_length(M);
		return( 0 );
	}
}

process_block( M )
unsigned long M[16];
{
	int	t;
	register	A, B, C, D, E;
	unsigned long	TEMP;

	/* copy M into the first 16 words of W */

	for( t=0; t<16; t++ ) W[t] = M[t];

	for( t=16; t<80; t++ )
	{
		W[t] = W[t-3] ^ W[t-8] ^ W[t-14] ^ W[t-16];
	}
	A = H0; B = H1; C = H2; D = H3; E = H4;

	if( verbose || trace )
	{
		if( blockcount == 0 )
		{
			printf("The initial hex values of {H[i]} are\n\n");
			dump_H_values(H0,H1,H2,H3,H4);
			printf("\n");
		}
	}
	if( trace )
	{
		blockcount++;
		printf("Start processing block %d.  ",blockcount);
		printf("The words of block %d are",blockcount);
		show_it( M );
		printf("The hex values of A,B,C,D,E after pass t of the ");
		printf("\"for t=0 to 79\" loop are\n\n");
		printf("\t   A\t    B\t     C\t      D\t       E\n\n");
	}

	for( t=0; t<20; t++ )
	{
		TEMP = (S(5,A)) + f0(B,C,D) + E + W[t] + K0;
		E = D; D = C; C = S(30,B); B = A; A = TEMP;
		if( trace ) dump_ABCDE( t,A,B,C,D,E );
	}
	for( t=20; t<40; t++ )
	{
		TEMP = (S(5,A)) + f1(B,C,D) + E + W[t] + K1;
		E = D; D = C; C = S(30,B); B = A; A = TEMP;
		if( trace ) dump_ABCDE( t,A,B,C,D,E );
	}
	for( t=40; t<60; t++ )
	{
		TEMP = (S(5,A)) + f2(B,C,D) + E + W[t] + K2;
		E = D; D = C; C = S(30,B); B = A; A = TEMP;
		if( trace ) dump_ABCDE( t,A,B,C,D,E );
	}
	for( t=60; t<80; t++ )
	{
		TEMP = (S(5,A)) + f3(B,C,D) + E + W[t] + K3;
		E = D; D = C; C = S(30,B); B = A; A = TEMP;
		if( trace ) dump_ABCDE( t,A,B,C,D,E );
	}
	H0 = H0 + A;
	H1 = H1 + B;
	H2 = H2 + C;
	H3 = H3 + D;
	H4 = H4 + E;
	
	if( trace )
	{
		printf("\nBlock %d has been processed.  ",blockcount);
		printf("The values of {H[i]} are\n\n");
		dump_H_values(H0,H1,H2,H3,H4);
		printf("\n");
	}
}

stuff_file_length(B)
unsigned int	*B;
{
	B[14] = sizebuf[0];
	B[15] = sizebuf[1];
}

update_sizebuf( bitcount )
int	bitcount;
{
	if( bitcount == 0 ) return(0);

	if( bitcount == 512 )
	{
		sizebuf[1] += 512;
		if( sizebuf[1] == 0 )
		{
			sizebuf[0]++;
		}
	}
	else	sizebuf[1] += bitcount;

	return(0);
}

print_results()
{
	if( verbose || trace )
		printf("Message digest =");
	print_hex( H0 );
	print_hex( H1 );
	print_hex( H2 );
	print_hex( H3 );
	print_hex( H4 );
	printf("\n");
	fflush(stdout);
}

/* sha special bitwise functions */

unsigned long	S(n,x)	/* circular left-shift n bits */
int	n;
unsigned long x;
{
	return( (x<<n) | (x>>(32-n)) );
}
unsigned long	f0(B,C,D)
unsigned long	B, C, D;
{
	return( ( B & C ) | ( ~B & D ) );
}

unsigned long	f1(B,C,D)
unsigned long	B, C, D;
{
	return( ( B^C^D ) );
}

unsigned long	f2(B,C,D)
unsigned long	B, C, D;
{
	return( ( B & C ) | ( B & D ) | ( C & D ) );
}

unsigned long	f3(B,C,D)
unsigned long	B, C, D;
{
	return( ( B^C^D ) );
}

print_bits( x, mode )
unsigned long x;
int	mode;
{
	int 	i;
	unsigned long mask = 0x80000000;
	unsigned long hold;

	if( mode == 2 ) hold = x;

	for( i=1; i<=32; i++ )
	{
		if( x & mask ) putchar( '1' );
		else putchar( '0' );
		if( ! (i%8) ) putchar( ' ' );
		x <<= 1;
	}
	if( mode == 2 )
		print_hex( hold );
	putchar( '\n' );
}

print_hex( x )
unsigned long	x;
{
	printf(" %08X",x);
}

dump_H_values( a,b,c,d,e )
unsigned long	a,b,c,d,e;
{
	printf("\t   H0\t    H1\t     H2\t      H3       H4\n\n       ");
	print_hex( a );
	print_hex( b );
	print_hex( c );
	print_hex( d );
	print_hex( e );
	printf("\n");
}

dump_ABCDE( pass,a,b,c,d,e )
int	pass;
unsigned long	a,b,c,d,e;
{
	printf("t = %2d:",pass);
	print_hex( a );
	print_hex( b );
	print_hex( c );
	print_hex( d );
	print_hex( e );
	printf("\n");
}

dump_sizebuf()
{
	printf("dumping sizebuf\n");
	printf("%08X ",sizebuf[0]);
	printf("%08X ",sizebuf[1]);
	printf("\n");
}

usage_exit(s)
char	*s;
{
	if( s[0] != '\0' ) printf("%s\n",s);
	printf("usage:  sha [-h][-v][-V][file]\n");
	exit(0);
}
